

import fnmatch

FILES = []


def populate(files):
    """Generate list of files from fixture data."""
    global FILES
    FILES = sorted(set(files))


def glob(pattern):
    return fnmatch.filter(FILES, pattern)



""" test gpath

   isort:skip_file
"""

import os
import sys
import mock
import unittest

SRC = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "src")
if SRC not in sys.path:
    sys.path.insert(0, SRC)

from ciopath.gpath import Path

sys.modules["glob"] = __import__("mocks.glob", fromlist=["dummy"])


class BadInputTest(unittest.TestCase):

    def test_empty_input(self):
        with self.assertRaises(ValueError):
            self.p = Path("")
            
class RootPath(unittest.TestCase):
    def test_root_path(self):
        self.p = Path("/")
        self.assertEqual(self.p.fslash(), "/")
        self.assertEqual(self.p.bslash(), "\\")

    def test_drive_letter_root_path(self):
        self.p = Path("C:\\")
        self.assertEqual(self.p.fslash(), "C:/")
        self.assertEqual(self.p.bslash(), "C:\\")


class SpecifyDriveLetterUse(unittest.TestCase):
    def test_remove_from_path(self):
        self.p = Path("C:\\a\\b\\c")
        self.assertEqual(self.p.fslash(with_drive=False), "/a/b/c")
        self.assertEqual(self.p.bslash(with_drive=False), "\\a\\b\\c")

    def test_remove_from_root_path(self):
        self.p = Path("C:\\")
        self.assertEqual(self.p.fslash(with_drive=False), "/")
        self.assertEqual(self.p.bslash(with_drive=False), "\\")




class AbsPosixPathTest(unittest.TestCase):
    def setUp(self):
        self.p = Path("/a/b/c")

    def test_fslash_out(self):
        self.assertEqual(self.p.fslash(), "/a/b/c")

    def test_win_path_out(self):
        self.assertEqual(self.p.bslash(), "\\a\\b\\c")


class AbsWindowsPathTest(unittest.TestCase):
    def setUp(self):
        self.p = Path("C:\\a\\b\\c")

    def test_fslash_out(self):
        self.assertEqual(self.p.fslash(), "C:/a/b/c")

    def test_win_path_out(self):
        self.assertEqual(self.p.bslash(), "C:\\a\\b\\c")

    # consider just testing on both platforms
    def test_os_path_out(self):
        with mock.patch("os.name", "posix"):
            self.assertEqual(self.p.os_path(), "C:/a/b/c")
        with mock.patch("os.name", "nt"):
            self.assertEqual(self.p.os_path(), "C:\\a\\b\\c")

class PathStringTest(unittest.TestCase):

    def test_path_emits_string_posix(self):
        input_file = "/path/to/thefile.jpg"
        p = Path(input_file)
        self.assertEqual(str(p) , input_file)

    def test_path_emits_string_with_drive(self):
        input_file = "C:/path/to/thefile.jpg"
        p = Path(input_file)
        self.assertEqual(str(p) , input_file)

    def test_path_emits_string_relative(self):
        input_file = "path/to/thefile.jpg"
        p = Path(input_file)
        self.assertEqual(str(p) , input_file)





class WindowsMixedPathTest(unittest.TestCase):
    def test_abs_in_fslash_out(self):
        self.p = Path("\\a\\b\\c/d/e")
        self.assertEqual(self.p.fslash(), "/a/b/c/d/e")

    def test_abs_in_bslash_out(self):
        self.p = Path("\\a\\b\\c/d/e")
        self.assertEqual(self.p.bslash(), "\\a\\b\\c\\d\\e")

    def test_letter_abs_in_fslash_out(self):
        self.p = Path("C:\\a\\b\\c/d/e")
        self.assertEqual(self.p.fslash(), "C:/a/b/c/d/e")

    def test_letter_abs_in_bslash_out(self):
        self.p = Path("C:\\a\\b\\c/d/e")
        self.assertEqual(self.p.bslash(), "C:\\a\\b\\c\\d\\e")


class MiscPathTest(unittest.TestCase):
    def test_many_to_single_backslashes_bslash_out(self):
        self.p = Path("C:\\\\a\\b///c")
        self.assertEqual(self.p.bslash(), "C:\\a\\b\\c")


class PathExpansionTest(unittest.TestCase):
    def setUp(self):
        self.env = {
            "HOME": "/users/joebloggs",
            "SHOT": "/metropolis/shot01",
            "DEPT": "texturing",
        }

    def test_posix_tilde_input(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("~/a/b/c")
            self.assertEqual(self.p.fslash(), "/users/joebloggs/a/b/c")

    def test_posix_var_input(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$SHOT/a/b/c")
            self.assertEqual(self.p.fslash(), "/metropolis/shot01/a/b/c")

    def test_posix_two_var_input(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$SHOT/a/b/$DEPT/c")
            self.assertEqual(self.p.fslash(),
                             "/metropolis/shot01/a/b/texturing/c")

    def test_windows_var_input(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$HOME\\a\\b\\c")
            self.assertEqual(self.p.bslash(),
                             "\\users\\joebloggs\\a\\b\\c")
            self.assertEqual(self.p.fslash(), "/users/joebloggs/a/b/c")

    def test_tilde_no_expand(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("~/a/b/c", no_expand=True)
            self.assertEqual(self.p.fslash(), "~/a/b/c")

    def test_posix_var_no_expand(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$SHOT/a/b/c", no_expand=True)
            self.assertEqual(self.p.fslash(), "$SHOT/a/b/c")

    def no_expand_variable_considered_relative(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$SHOT/a/b/c", no_expand=True)
            self.assertTrue(self.p.relative)
            self.assertFalse(self.p.absolute)

    def expanded_variable_considered_absolute(self):
        with mock.patch.dict("os.environ", self.env):
            self.p = Path("$SHOT/a/b/c", no_expand=False)
            self.assertFalse(self.p.relative)
            self.assertTrue(self.p.absolute)




class PathContextExpansionTest(unittest.TestCase):
    def setUp(self):

        self.env = {
            "HOME": "/users/joebloggs",
            "SHOT": "/metropolis/shot01",
            "DEPT": "texturing",
        }

        self.context = {
            "HOME": "/users/janedoe",
            "FOO": "fooval",
            "BAR_FLY1_": "bar_fly1_val",
            "ROOT_DIR": "/some/root",
        }

    def test_path_replaces_context(self):
        self.p = Path("$ROOT_DIR/thefile.jpg", context=self.context)
        self.assertEqual(self.p.fslash(), "/some/root/thefile.jpg")

    def test_path_replaces_multiple_context(self):
        self.p = Path("$ROOT_DIR/$BAR_FLY1_/thefile.jpg", context=self.context)
        self.assertEqual(self.p.fslash(),
                         "/some/root/bar_fly1_val/thefile.jpg")

    def test_path_context_overrides_env(self):
        self.p = Path("$HOME/thefile.jpg", context=self.context)
        self.assertEqual(self.p.fslash(), "/users/janedoe/thefile.jpg")

    def test_path_leave_unknown_variable_in_tact(self):
        self.p = Path("$ROOT_DIR/$BAR_FLY1_/$FOO/thefile.$F.jpg",
                      context=self.context)
        self.assertEqual(
            self.p.fslash(), "/some/root/bar_fly1_val/fooval/thefile.$F.jpg"
        )


class PathLengthTest(unittest.TestCase):
    def test_len_with_drive_letter(self):
        self.p = Path("C:\\aaa\\bbb/c")
        self.assertEqual(len(self.p), 12)

    def test_len_with_no_drive_letter(self):
        self.p = Path("\\aaa\\bbb/c")
        self.assertEqual(len(self.p), 10)

    def test_depth_with_drive_letter(self):
        self.p = Path("C:\\aaa\\bbb/c")
        self.assertEqual(self.p.depth, 3)

    def test_depth_with_no_drive_letter(self):
        self.p = Path("\\aaa\\bbb/c")
        self.assertEqual(self.p.depth, 3)

    def test_depth_with_literal_rel_path(self):
        self.p = Path("aaa\\bbb/c")
        self.assertEqual(self.p.depth, 3)


class PathCollapseDotsTest(unittest.TestCase):
    def test_path_collapses_single_dot(self):
        p = Path("/a/b/./c")
        self.assertEqual(p.fslash(), "/a/b/c")

    def test_path_collapses_double_dot(self):
        p = Path("/a/b/../c")
        self.assertEqual(p.fslash(), "/a/c")

    def test_path_collapses_many_single_dots(self):
        p = Path("/a/b/./c/././d")
        self.assertEqual(p.fslash(), "/a/b/c/d")

    def test_path_collapses_many_consecutive_double_dots(self):
        p = Path("/a/b/c/../../d")
        self.assertEqual(p.fslash(), "/a/d")

    def test_path_collapses_many_non_consecutive_double_dots(self):
        p = Path("/a/b/c/../../d/../e/f/../g")
        self.assertEqual(p.fslash(), "/a/e/g")

    def test_path_collapses_many_non_consecutive_mixed_dots(self):
        p = Path("/a/./b/c/../.././d/../././e/f/../g/./")
        self.assertEqual(p.fslash(), "/a/e/g")
        self.assertEqual(p.depth, 3)

    def test_path_collapses_to_root(self):
        p = Path("/a/b/../../")
        self.assertEqual(p.fslash(), "/")
        self.assertEqual(p.depth, 0)

    def test_raise_when_collapse_too_many_dots(self):
        with self.assertRaises(ValueError):
            Path("/a/b/../../../")


class PathComponentsTest(unittest.TestCase):
    def test_path_gets_tail(self):
        p = Path("/a/b/c")
        self.assertEqual(p.tail, "c")

    def test_path_gets_none_when_no_tail(self):
        p = Path("/")
        self.assertEqual(p.tail, None)

    def test_path_ends_with(self):
        p = Path("/a/b/cdef")
        self.assertTrue(p.endswith("ef"))

    def test_path_not_ends_with(self):
        p = Path("/a/b/cdef")
        self.assertFalse(p.endswith("eg"))


class RelativePathTest(unittest.TestCase):
    def test_rel_path_does_not_raise(self):
        p = Path("a/b/c")
        self.assertEqual(p.fslash(), "a/b/c")

 
class EqualityTests(unittest.TestCase):
    def test_paths_equal(self):
        p1 = Path("a/b/c")
        p2 = Path("a/b/c")
        self.assertTrue(p1 == p2)

    def test_same_object_equal(self):
        p1 = Path("a/b/c")
        self.assertTrue(p1 == p1)

    def test_different_paths_equal_false(self):
        p1 = Path("a/b/c")
        p2 = Path("a/b/d")
        self.assertFalse(p1 == p2)

    def test_paths_not_equal(self):
        p1 = Path("a/b/c")
        p2 = Path("a/b/d")
        self.assertTrue(p1 != p2)


if __name__ == "__main__":
    unittest.main()



from future.utils import python_2_unicode_compatible

import os

import re

RX_LETTER = re.compile(r"^([a-zA-Z]):[\\/]+")
RX_DOLLAR_VAR = re.compile(r"\$([A-Za-z][A-Z,a-z0-9_]+)")


def _expand_context(path, context):
    result = path
    if context:
        for match in RX_DOLLAR_VAR.finditer(path):
            key = match.group(1)
            replacement = context.get(key)
            if replacement is not None:
                result = result.replace("${}".format(key), replacement)
    return result


def _normalize_dots(components):
    currentdir = "."
    parentdir = ".."
    result = []
    for c in components:
        if c == currentdir:
            pass
        elif c == parentdir:
            if not len(result):
                raise ValueError("Can't resolve components due to '..' overflow.")
            del result[-1]
        else:
            result.append(c)
    return result


class Path(object):
    def __init__(self, path, **kw):
        """Initialize a generic absolute path.

        If path is a list, then each element will be a component of the path. 
        If it's a string then expand context variables.
        Also expand, env vars and user unless explicitly told not to with the
        no_expand option. 
        """
        self._drive_letter = None

        if not path:
            raise ValueError("Empty path")

        if isinstance(path, list):
            ipath = path[:]
            self._drive_letter = ipath.pop(0)[0] if RX_LETTER.match(ipath[0]) else None
            self._components = _normalize_dots(ipath)
            self._absolute = True
        else:
            context = kw.get("context")
            if context:
                path = _expand_context(path, context)

            if not kw.get("no_expand", False):
                path = os.path.expanduser(os.path.expandvars(path))

            match = RX_LETTER.match(path)
            # print "match", match, match.group(1) if match else None
            if match:
                self._drive_letter = match.group(1)
                path = path.split(":", 1)[1]
    
            self._absolute = path[0] in ["/", "\\"]

            self._components = _normalize_dots(
                [s for s in re.split("/|\\\\", path) if s]
            )

        self._depth = len(self._components)

    def _construct_path(self, sep, with_drive_letter=True):
        """Reconstruct path for given path sep."""
        result = sep.join(self._components)
        if self._absolute:
            result = "{}{}".format(sep, result)
            if with_drive_letter and self._drive_letter:
                result = "{}:{}".format(self._drive_letter, result)
        return result

    def fslash(self, **kw):
        """Path with forward slashes. Can include drive letter."""
        with_drive_letter = kw.get("with_drive", True)
        return self._construct_path("/", with_drive_letter)

    def bslash(self, **kw):
        """Path with back slashes. Can include drive letter."""
        with_drive_letter = kw.get("with_drive", True)
        return self._construct_path("\\", with_drive_letter)

    def os_path(self, **kw):
        """Path with slashes for current os. Can include drive letter."""
        with_drive = kw.get("with_drive", True)
        if os.name == "nt":
            return self.bslash(with_drive=with_drive)
        return self.fslash(with_drive=with_drive)

    def __str__(self):
        """Same as fslash, with_drive=True"""
        return self._construct_path("/", True)

    def startswith(self, path):
        return self.fslash().startswith(path.fslash())

    def endswith(self, suffix):
        return self.fslash().endswith(suffix)

    def __len__(self):
        return len(self.fslash())

    def __eq__(self, rhs):
        if not isinstance(rhs, Path):
            raise NotImplementedError
        return self.fslash() == rhs.fslash()

    def __ne__(self, rhs):
        return not (self == rhs)

    @property
    def depth(self):
        return self._depth

    @property
    def drive_letter(self):
        return self._drive_letter or ""

    @property
    def absolute(self):
        return self._absolute

    @property
    def relative(self):
        return not self._absolute

    @property
    def components(self):
        return self._components or []

    @property
    def all_components(self):
        if self.drive_letter:
            return ["{}:".format(self.drive_letter)] + self.components
        else:
            return self.components

    @property
    def tail(self):
        return self._components[-1] if self._components else None



import os
import re
from itertools import takewhile
import glob

from ciopath.gpath import Path
from builtins import object

GLOBBABLE_REGEX = re.compile(r"\*|\?|\[")


class PathList(object):
    """A list of files with lazy deduplication.

    Every time the private entrylist is accessed with an accessor such
    as len or iter, the _clean flag is checked. If it's dirty we
    deduplicate and set it clean. It becomes dirty again if files are
    added. Since adding a single directory can cause many contained
    files to be removed during deduplication, we set the next iterator
    to zero.
    """

    def __init__(self, *paths):
        """Initialize."""
        self._entries = []
        self._clean = False
        self._current = 0
        self.add(*paths)

    def add(self, *paths):
        """Add one or more files.

        Duplicate files and directories that contain other files may be
        added and no deduplication will happen at this time.
        """

        for path in paths:
            self._add_one(path)

    def _add_one(self, path):
        """Add a single file.

        Note that when an element is added, it may cause the list to
        change next time it is deduplicated, which includes getting
        shorter. This could happen if a containing directory is added.
        Therefore we have to set the peg position to zero.
        """

        if not type(path).__name__ == "Path":
            path = Path(path)
        self._entries.append(path)
        self._clean = False
        self._current = 0

    def remove(self, *paths):
        """
        Replace the underlying list with a filtered list. 

        No deduplication happens yet and the list is marked dirty.
        """
        removals = PathList(*paths)
        result = [p for p in self._entries if p not in removals]
        self._entries = result
        self._clean = False
        self._current = 0

    def _deduplicate(self):
        """Deduplicate if it has become dirty.

        Algorithm explained at:
        https://stackoverflow.com/questions/49478361 .
        """
        if self._clean:
            return

        sorted_entries = sorted(
            self._entries, key=lambda entry: (entry.depth, -len(entry.tail))
        )

        self._entries = []
        for entry in sorted_entries:
            if any(entry.startswith(p) for p in self._entries):
                continue
            self._entries.append(entry)
        self._clean = True

    def __contains__(self, key):
        if not isinstance(key, Path):
            key = Path(key)
        return key in self._entries

    def common_path(self):
        """Find the common path among entries.

        This is useful for determining output directory when many renders are
        rendering to different places.

        In the case where only single path exists, it is not possible to tell
        from its name whether it is a file or directory. We don't want this
        method to touch the filesystem, that should be someone else's problem. A
        trailing slash would be a hint, but the absence of a trailing slash does
        not mean its a regular file. Therefore, in the case of a single file we
        return it AS-IS and the caller can then stat to find out for sure.

        If no files exist return None.

        If the filesystem root is the common path, return root path, which is
        not entirely correct on windows with drive letters.
        """
        if not self._entries:
            return None

        absolute =  list(self._entries)[0].absolute

        def _all_the_same(rhs):
            return all(n == rhs[0] for n in rhs[1:])

        levels = zip(*[p.all_components for p in self._entries])

        common = [x[0] for x in takewhile(_all_the_same, levels)]

        if not len(common):
            return Path("/")
        elif  common[0].endswith(":"):
            return Path("/".join(common))
        elif absolute: 
            return Path("/"+"/".join(common))
        return Path("/".join(common))

       

    def glob(self):
        """Glob expansion for entries containing globbable characters.

        We don't simply glob every entry since that would remove entries
        that don't yet exist. And we can't just rely on zero glob
        results because it may have been a legitimate zero result if it
        was globbable but matched nothing. So we test for glob
        characters (*|?|[) to determine whether to attempt a glob.

        However, if it looks like a glob, but the glob library can't
        handle it, then we have to assume it really is a filename with 
        glob-like characters, and then we just add the literal path 
        unchanged. See the test: test_ignore_invalid_glob().
        """
        self._deduplicate()
        result = []
        for entry in self._entries:
            pp = entry.fslash()
            if GLOBBABLE_REGEX.search(pp):
                try:
                    globs = glob.glob(entry.fslash())
                    result += globs
                except re.error:
                    result.append(pp)
            else:
                result.append(pp)
        self._entries = [Path(g) for g in result]
        self._clean = False
        self._current = 0

    def __iter__(self):
        """Get an iterator to entries.

        Deduplicate just in time.
        """
        self._deduplicate()
        return iter(self._entries)

    def __next__(self):
    # def next(self):
        """Get the next element.

        Deduplicate just in time.
        """
        self._deduplicate()
        if self._current >= len(self._entries):
            raise StopIteration
        else:
            prev = self._current
            self._current += 1
            return self._entries[prev]

    def __len__(self):
        """Get the size of the entry list.

        Deduplicate just in time.
        """
        self._deduplicate()
        return len(self._entries)

    def remove_missing(self):
        missing = PathList()
        for path in self._entries:
            pp = path.fslash()
            if GLOBBABLE_REGEX.search(pp):
                continue
            if not os.path.exists(pp):
                missing.add(path)
        if  missing:
            self.remove(*missing)



""" test gpathlist

   isort:skip_file
"""
 
import sys
import os
import unittest
import mock
from mock import patch

SRC = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "src")
if SRC not in sys.path:
    sys.path.insert(0, SRC)


# Make sure the mocked version of glob is in the sys modules
sys.modules["glob"] = __import__("mocks.glob", fromlist=["dummy"])

# Import gpath_list and anything else that may use glob which will NOT import
# the real glob because it's already here.
from ciopath.gpath_list import PathList
from ciopath.gpath import Path
# from cioseq.sequence import Sequence

# Now import glob because we need to populate it.
import glob # isort:skip

class PathListTest(unittest.TestCase):
    def setUp(self):
        self.env = {
            "HOME": "/users/joebloggs",
            "SHOT": "/metropolis/shot01",
            "DEPT": "texturing",
        }

        self.other_files_on_disk = [
         '/other/file.0001.exr',
         '/other/file.0002.exr',
         '/other/file.0003.exr',
         '/other/file.0004.exr',
         '/other/file.0005.exr',
         '/other/file.0006.exr',
         '/other/file.0007.exr',
         '/other/file.0008.exr',
         '/other/file.0009.exr',
         '/other/file.0010.exr',
         '/other/file.0011.exr',
         '/other/file.0012.exr',
         '/other/file.0013.exr',
         '/other/file.0014.exr',
         '/other/file.0015.exr',
         '/other/file.0016.exr',
         '/other/file.0017.exr',
         '/other/file.0018.exr',
         '/other/file.0019.exr',
         '/other/file.0020.exr']

        self.some_files_on_disk = [
         '/some/file.0001.exr',
         '/some/file.0002.exr',
         '/some/file.0003.exr',
         '/some/file.0004.exr',
         '/some/file.0005.exr',
         '/some/file.0006.exr',
         '/some/file.0007.exr',
         '/some/file.0008.exr',
         '/some/file.0009.exr',
         '/some/file.0010.exr',
         '/some/file.0011.exr',
         '/some/file.0012.exr',
         '/some/file.0013.exr',
         '/some/file.0014.exr',
         '/some/file.0015.exr',
         '/some/file.0016.exr',
         '/some/file.0017.exr',
         '/some/file.0018.exr',
         '/some/file.0019.exr',
         '/some/file.0020.exr']



    def test_init_empty(self):
        d = PathList()
        self.assertEqual(list(d), [])

    def test_init_with_args(self):
        d = PathList("/a/file2", "/a/file3")
        self.assertEqual(len(d), 2)

    def test_adds_paths(self):
        d = PathList()
        d.add(Path("/a/file1"), Path("/a/file2"))
        self.assertEqual(len(d), 2)

    def test_adds_strings(self):
        d = PathList()
        d.add("/a/file1", "/a/file2")
        self.assertEqual(len(d), 2)

    def test_adds_mix(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", Path("/a/file3"))
        self.assertEqual(len(d), 3)


    # remove
    def test_removes_string(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3")
        d.remove("/a/file2")
        self.assertEqual(len(d), 2)

    def test_removes_path(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3")
        d.remove(Path("/a/file2"))
        self.assertEqual(len(d), 2)
        self.assertEqual( list(d)[1],  Path("/a/file3"))

    def test_removes_many(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3")
        d.remove("/a/file2", "/a/file3")
        self.assertEqual(len(d), 1)

    def test_removes_paths_when_list_contains_duplicates(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3", "/a/file2")
        d.remove("/a/file1")
        self.assertEqual(len(d), 2)

    def test_removes_duplicate_when_list_contains_duplicates(self):
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3", "/a/file2", "/a/file3")
        d.remove("/a/file2")
        self.assertEqual(len(d), 2)
 
    def test_does_not_deduplicate_internally_on_removal(self):
        # deduplication is lazy - happens on access
        d = PathList()
        d.add("/a/file1", "/a/file2", "/a/file3", "/a/file2", "/a/file3")
        d.remove("/a/file2")
        self.assertEqual(len(d._entries), 3)


    # just want to make sure expansion works here
    # even though it's tested in gpath_test
    def test_expand_tilde(self):
        with mock.patch.dict("os.environ", self.env):
            d = PathList()
            d.add("~/file1", "~/file2")

            self.assertIn("/users/joebloggs/file1", d)

    def test_expand_envvar(self):
        with mock.patch.dict("os.environ", self.env):
            d = PathList()
            d.add("$SHOT/file1", "$HOME/file2")
            self.assertIn("/metropolis/shot01/file1", d)
            self.assertIn("/users/joebloggs/file2", d)

    def test_dedup_same_paths(self):
        d = PathList()
        d.add(Path("/file1"), Path("/file2"), Path("/file2"))
        self.assertEqual(len(d), 2)
        self.assertIn(Path("/file1"), d)
        self.assertIn(Path("/file2"), d)

    def test_dedup_same_strings(self):
        d = PathList()
        d.add("/file1", "/file2", "/file2")
        self.assertEqual(len(d), 2)
        self.assertIn("/file1", d)
        self.assertIn("/file2", d)

    def test_dedup_contained_file(self):
        d = PathList()
        d.add("/dir1/", "/dir1/file1", "/dir2/file1", "/dir3/file2")
        self.assertEqual(len(d), 3)

    def test_dedup_dirtied_on_add(self):
        d = PathList()
        d.add("/file1")
        self.assertFalse(d._clean)

    def test_dedup_cleaned_on_access_iter(self):
        d = PathList()
        d.add("/file1")
        ls = list(d)
        self.assertTrue(d._clean)

    def test_dedup_cleaned_on_access_len(self):
        d = PathList()
        d.add("/file1")
        ls = len(d)
        self.assertTrue(d._clean)

    def test_dedup_cleaned_on_access_next(self):
        d = PathList()
        d.add("/file1", "/file2", "/file3")
        n = next(d)
        self.assertTrue(d._clean)

    def test_next(self):
        d = PathList()
        d.add("/file1", "/file2", "/file3")
        self.assertEqual(next(d), Path("/file1"))
        self.assertEqual(next(d), Path("/file2"))

    def test_next_fails_after_last(self):
        d = PathList()
        d.add("/file1", "/file2", "/file3")
        next(d)
        next(d)
        next(d)
        with self.assertRaises(StopIteration):
            next(d)

    def test_next_reset_after_add(self):
        d = PathList()
        d.add("/file1", "/file2", "/file3")
        next(d)
        next(d)
        d.add("/file4")
        self.assertEqual(next(d), Path("/file1"))

    def test_common_path_when_common_prefix_in_filename(self):
        d = PathList()
        files = [
            "/users/joebloggs/tmp/dissention/perfect",
            "/users/joebloggs/tmp/disagreement/crimson",
            "/users/joebloggs/tmp/diatribe/belew",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp"))

    def test_common_path(self):
        d = PathList()
        files = [
            "/users/joebloggs/tmp/foobar/test",
            "/users/joebloggs/tmp/baz/fripp",
            "/users/joebloggs/tmp/elephant/corner",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp"))

    def test_common_path_when_one_path_is_the_common_path(self):
        d = PathList()
        files = [
            "/users/joebloggs/tmp",
            "/users/joebloggs/tmp/bolly/operation",
            "/users/joebloggs/tmp/stay/go",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp"))

    def test_common_path_when_lowest_path_is_the_common_path(self):
        d = PathList()
        files = [
            "/users/joebloggs/tmp/foo.txt",
            "/users/joebloggs/tmp/modelman.jpg",
            "/users/joebloggs/tmp/ration.cpp",
            "/users/joebloggs/tmp/bill.project",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp"))

    def test_common_path_drive_letter(self):
        d = PathList()
        files = [
            "C:/users/joebloggs/hello/foo.txt",
            "C:/users/joebloggs/tmp/modelman.jpg",
            "C:/users/joebloggs/tmp/ration.cpp",
            "C:/users/joebloggs/tmp/bill.project",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("C:/users/joebloggs"))

    # This is not right. There is no common path if drive letters
    # are involved. Need to revisit, and hope that in the
    # meantime  no one renders to two different filesystems in the
    # same render job.
    def test_common_different_drive_letter(self):
        d = PathList()
        files = [
            "D:/users/joebloggs/tmp/foo.txt",
            "D:/users/joebloggs/tmp/modelman.jpg",
            "C:/users/joebloggs/tmp/ration.cpp",
            "C:/users/joebloggs/tmp/bill.project",
        ]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/"))

    def test_common_path_when_single_path(self):
        d = PathList()
        files = ["/users/joebloggs/tmp/foo.txt"]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp/foo.txt"))

    def test_common_path_when_duplicate_entries_of_single_path(self):
        d = PathList()
        files = ["/users/joebloggs/tmp/foo.txt",
                 "/users/joebloggs/tmp/foo.txt"]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/users/joebloggs/tmp/foo.txt"))

    def test_common_path_is_none_when_no_entries(self):
        d = PathList()
        self.assertIsNone(d.common_path())

    def test_common_path_is_slash_when_root(self):
        d = PathList()
        files = ["/users/joebloggs/tmp/foo.txt", "/dev/joebloggs/tmp/foo.txt"]
        d.add(*files)
        self.assertEqual(d.common_path(), Path("/"))

    def test_glob_when_files_match_with_asterisk(self):
        glob.populate(self.some_files_on_disk)
        d = PathList()
        file = "/some/file.*.exr"
        d.add(file)
        d.glob()
        self.assertEqual(len(d), 20)

    def test_glob_when_files_match_with_question_mark(self):
        glob.populate(self.some_files_on_disk)
        d = PathList()
        file = "/some/file.00?0.exr"
        d.add(file)
        d.glob()
        self.assertEqual(len(d), 2)

    def test_glob_when_files_match_with_range(self):
        glob.populate(self.some_files_on_disk)
        d = PathList()
        file = "/some/file.000[0-9].exr"
        d.add(file)
        d.glob()
        self.assertEqual(len(d), 9)

    def test_glob_dedups_when_many_files_match(self):
        glob.populate(self.some_files_on_disk)
        d = PathList()
        files = ["/some/file.*.exr", "/some/*.exr"]
        d.add(*files)
        d.glob()
        self.assertEqual(len(d), 20)

    def test_glob_when_files_dont_match(self):
        glob.populate(self.other_files_on_disk)
        d = PathList()
        file = "/some/file.*.exr"
        d.add(file)
        d.glob()
        self.assertEqual(len(d), 0)

    def test_unpacking(self):
        d = PathList()
        d.add(Path("/a/file1"), Path("/a/file2"))
        a, b = d
        self.assertEqual(type(a), Path)

    def test_glob_leaves_non_existent_unglobbable_entries_untouched(self):
        glob.populate(self.some_files_on_disk[:3])
        d = PathList()
        d.add("/some/file.*.exr", "/other/file1.exr", "/other/file2.exr")
        d.glob()
        self.assertEqual(len(d), 5)

    def test_ignore_invalid_glob(self):
        bad_glob = "/path/to/Model[b-a]"
        glob.populate(bad_glob)
        d = PathList()
        d.add(bad_glob)
        d.glob()
        self.assertEqual(list(d)[0].fslash(), bad_glob)

    def test_ignore_invalid_nonexistent_glob(self):
        bad_glob = "/path/to/Model[b-a]"
        d = PathList()
        d.add(bad_glob)
        d.glob()
        self.assertEqual(list(d)[0].fslash(), bad_glob)
 
    def test_empty_list_is_falsy(self):
        self.assertFalse(PathList())

    def test_populated_list_is_truthy(self):
        self.assertTrue(PathList("/file"))


class MissingFilesTest(unittest.TestCase):

    @staticmethod
    def side_effect(arg):
        if "missing" in arg: 
            return False
        else:
            return True

    def setUp(self):

        patcher = patch('os.path.exists')
        self.mock_exists = patcher.start()
        self.mock_exists.side_effect = MissingFilesTest.side_effect
        self.addCleanup(patcher.stop)


    def test_remove_no_missing_file(self):
        d = PathList()
        files = ["/tmp/foo.txt", "/tmp/bar.txt"]
        d.add(*files)
        self.assertEqual(len(d), 2)
        d.remove_missing()
        self.assertEqual(len(d), 2)

    def test_remove_one_missing_file(self):
        d = PathList()
        files = ["/tmp/missing.txt", "/tmp/foo.txt", "/tmp/bar.txt"]
        d.add(*files)
        self.assertEqual(len(d), 3)
        d.remove_missing()
        self.assertEqual(len(d), 2)

    def test_remove_many_missing_file(self):
        d = PathList()
        files = ["/tmp/missing.txt", "/tmp/foo.txt", "/tmp/bar.txt", "/tmp/missing2.txt", "/tmp/missing3.txt"]
        d.add(*files)
        self.assertEqual(len(d), 5)
        d.remove_missing()
        self.assertEqual(len(d), 2)

    def test_remove_when_all_files_missing(self):
        d = PathList()
        files = ["/tmp/missing.txt", "/tmp/missing2.txt", "/tmp/missing3.txt"]
        d.add(*files)
        self.assertEqual(len(d), 3)
        d.remove_missing()
        self.assertFalse(d)

    def test_remove_missing_when_dups_given(self):
        d = PathList()
        files = ["/tmp/missing", "/tmp/foo", "/tmp/bar",  "/tmp/foo", "/tmp/missing2", "/tmp/missing"]
        d.add(*files)
        self.assertEqual(len(d), 4)
        d.remove_missing()
        self.assertEqual(len(d), 2)


    def test_dont_remove_globbable_files(self):
        d = PathList()
        files = ["/tmp/foo","/tmp/missing*", "/tmp/missing.[0-9]", "/tmp/missing.????.exr"]
        d.add(*files)
        self.assertEqual(len(d), 4)
        d.remove_missing()
        self.assertEqual(len(d), 4)

if __name__ == "__main__":
    unittest.main()
